/*
 * Copyright 2010-2024 OpenXcom Developers.
 *
 * This file is part of OpenXcom.
 *
 * OpenXcom is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OpenXcom is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenXcom.  If not, see <http://www.gnu.org/licenses/>.
 */

 // throw exceptions more often
//#define RYML_DEFAULT_CALLBACK_USES_EXCEPTIONS
//#define NDEBUG
//#define C4_NO_DEBUG_BREAK
// when using a single header file for the whole rapidayaml, this needs to be defined once, in one of the cpp files
// #define RYML_SINGLE_HDR_DEFINE_NOW
// include extra functionality for interaction with std types
//#define _RYML_STD_HPP_

//#include <ryml.hpp>
#include "Yaml.h"
#include <string>
#include <fstream>

namespace OpenXcom
{

namespace YAML
{


YamlString::YamlString(std::string yamlString) : yaml(yamlString) {}

YamlNodeReader::YamlNodeReader(const ryml::ConstNodeRef& node)
	: _node(node), _invalid(node.invalid()), _index(nullptr)
{
}

YamlNodeReader::YamlNodeReader(const ryml::ConstNodeRef& node, bool useIndex)
	: _node(node), _invalid(node.invalid()), _index(nullptr)
{
	if (!useIndex)
		return;
	// build and use an index to avoid [] operator's O(n) complexity
	_index = new std::unordered_map<ryml::csubstr, ryml::id_type>();
	_index->reserve(_node.num_children());
	for (const ryml::ConstNodeRef& childNode : _node.children())
		(*_index)[childNode.key()] = childNode.id();
}

YamlNodeReader::YamlNodeReader()
	: _node(ryml::ConstNodeRef(nullptr, ryml::NONE)), _invalid(true), _index(nullptr)
{
}

YamlNodeReader::~YamlNodeReader()
{
	if (_index)
		delete _index;
}

const YamlNodeReader YamlNodeReader::useIndex() const
{
	return YamlNodeReader(_node, true);
}

std::vector<char> YamlNodeReader::readValBase64() const
{
	//first run figure out the decoded length
	size_t len = _node.deserialize_val(ryml::fmt::base64(ryml::substr()));
	std::vector<char> decodedData(len);
	ryml::blob buf(decodedData.data(), len);
	//deserialize
	_node.deserialize_val(ryml::fmt::base64(buf));
	return decodedData;
}

size_t YamlNodeReader::childrenCount() const
{
	if (_invalid)
		return 0;
	return _node.num_children();
}

ryml::ConstNodeRef YamlNodeReader::getChildNode(ryml::csubstr key) const
{
	if (_invalid)
		return ryml::ConstNodeRef(_node.tree(), ryml::NONE);
	if (!_index)
		return _node.find_child(key);
	if (!_index->count(key))
		return ryml::ConstNodeRef(_node.tree(), ryml::NONE);
	return _node.tree()->cref(_index->at(key));
}

std::vector<YamlNodeReader> YamlNodeReader::children() const
{
	if (_invalid)
		return std::vector<YamlNodeReader>();
	std::vector<YamlNodeReader> children;
	children.reserve(_node.num_children());
	for (const ryml::ConstNodeRef& child : _node.cchildren())
		children.emplace_back(child);
	return children;
}

bool YamlNodeReader::isValid() const
{
	return !_invalid;
}

bool YamlNodeReader::isMap() const
{
	return _node.is_map();
}

bool YamlNodeReader::isSeq() const
{
	return _node.is_seq();
}

bool YamlNodeReader::hasVal() const
{
	return _node.has_val();
}

const YamlString YamlNodeReader::emit() const
{
	return YamlString(ryml::emitrs_yaml<std::string>(_node));
}

void YamlNodeReader::emit(std::string fullFilePath) const
{
	ryml::substr buf;
	ryml::emit_yaml(_node, buf);
	std::ofstream outFile(fullFilePath, std::ios::binary | std::ios::out);
	if (!outFile) { throw std::runtime_error("Unable to open file for writing."); }
	outFile.write(buf.str, buf.size());
	outFile.close();
}

const YamlNodeReader YamlNodeReader::operator[](ryml::csubstr key) const
{
	return YamlNodeReader(getChildNode(key));
}

const YamlNodeReader YamlNodeReader::operator[](size_t pos) const
{
	return YamlNodeReader(_node.child(pos));
}

YamlNodeReader::operator bool() const
{
	return !_invalid;
}

YamlRootNodeReader::YamlRootNodeReader() : YamlNodeReader(), _tree()
{
}

YamlRootNodeReader::YamlRootNodeReader(std::string fullFilePath, bool onlyInfoHeader) : YamlNodeReader(), _tree(new ryml::Tree())
{
	size_t size;
	//char* data = onlyInfoHeader ? CrossPlatform::getYamlSaveHeaderRaw(fullPath, &size) : CrossPlatform::readFileRaw(fullPath, &size);

	std::ifstream file(fullFilePath, std::ios::binary | std::ios::ate);
	if (!file) { throw std::runtime_error("Could not open file"); }
	size = (size_t)file.tellg();
	file.seekg(0, std::ios::beg);
	char* data = (char*)malloc(size);
	std::string buffer(size, '\0');
	file.read(data, size);
	file.close();

	_tree->reserve((size / 16)); //about 1 node every 16 bytes is a good estimate to minimize resizing
	ryml::parse_in_arena(ryml::csubstr(data, size), _tree);
	_node = _tree->crootref();
	_invalid = _node.invalid();
	//SDL_free(data);
	free(data);
}

YamlRootNodeReader::YamlRootNodeReader(char* data, size_t size) : YamlNodeReader(), _tree(new ryml::Tree())
{
	_tree->reserve((size / 16));
	ryml::parse_in_arena(ryml::csubstr(data, size), _tree);
	free(data);
	_node = _tree->crootref();
	_invalid = _node.invalid();
}

YamlRootNodeReader::YamlRootNodeReader(const YamlString& yamlString) : YamlNodeReader(), _tree(new ryml::Tree())
{
	_tree->reserve((yamlString.yaml.length() / 16));
	ryml::parse_in_arena(ryml::to_csubstr(yamlString.yaml), _tree);
	_node = _tree->crootref();
	_invalid = _node.invalid();
}

YamlRootNodeReader::~YamlRootNodeReader()
{
	delete _tree;
}

YamlNodeReader YamlRootNodeReader::sansRoot()
{
	return YamlNodeReader(_node);
}

YamlNodeWriter::YamlNodeWriter(ryml::NodeRef node) : _node(node)
{
}

YamlNodeReader YamlNodeWriter::toReader()
{
	return YamlNodeReader(_node);
}

YamlNodeWriter YamlNodeWriter::write()
{
	return YamlNodeWriter(_node.append_child());
}

YamlNodeWriter YamlNodeWriter::operator[](ryml::csubstr key)
{
	return YamlNodeWriter(_node.append_child({ryml::KEY, key}));
}

YamlNodeWriter YamlNodeWriter::writeBase64(ryml::csubstr key, char* data, size_t size)
{
	return YamlNodeWriter(_node.append_child({ ryml::KEY, key }) << ryml::fmt::base64(ryml::csubstr(data, size)));
}

void YamlNodeWriter::setAsMap()
{
	_node |= ryml::MAP;
}

void YamlNodeWriter::setAsSeq()
{
	_node |= ryml::SEQ;
}

void YamlNodeWriter::setFlowStyle()
{
	_node |= ryml::FLOW_SL;
}

void YamlNodeWriter::setBlockStyle()
{
	_node |= ryml::BLOCK;
}

ryml::csubstr YamlNodeWriter::saveString(const std::string& string)
{
	return _node.tree()->to_arena(string);
}

YamlString YamlNodeWriter::emit()
{
	return YamlString(ryml::emitrs_yaml<std::string>(_node));
}

YamlRootNodeWriter::YamlRootNodeWriter() :  YamlNodeWriter(_node), _tree(new ryml::Tree())
{
	_node = _tree->rootref();
}

YamlRootNodeWriter::YamlRootNodeWriter(size_t bufferCapacity) : YamlNodeWriter(_node), _tree(new ryml::Tree(bufferCapacity / 16, bufferCapacity))
{
	_node = _tree->rootref();
}

YamlRootNodeWriter::~YamlRootNodeWriter()
{
	delete _tree;
}

YamlNodeWriter YamlRootNodeWriter::sansRoot()
{
	return YamlNodeWriter(_node);
}

}

}
