#pragma once
/*
 * Copyright 2010-2024 OpenXcom Developers.
 *
 * This file is part of OpenXcom.
 *
 * OpenXcom is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OpenXcom is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenXcom.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma warning(push)
#pragma warning(disable : 4668 6011 6255 6293 6386 26439 26495 26498 26819)
#include <ryml.hpp>
#include <ryml_std.hpp>
#pragma warning(pop)
#include <unordered_map>
#include <fstream>

//hash function for ryml::csubstr for unordered_map -> same as std::hash<std::string>
template <> struct std::hash<ryml::csubstr>{ std::size_t operator()(const ryml::csubstr& k) const {	return _Hash_array_representation(k.str, k.len); } };

namespace OpenXcom
{

// Deserialization template for enums. We have to overload from_chars() instead of read(), because read() already has a template for all types -> template conflict
template <typename EnumType>
typename std::enable_if<std::is_enum<EnumType>::value, bool>::type inline from_chars(ryml::csubstr buf, EnumType* v) noexcept
{
	return ryml::atoi(buf, (int*)v);
}

template <typename EnumType>
typename std::enable_if<std::is_enum<EnumType>::value, size_t>::type inline to_chars(ryml::substr buf, EnumType v) noexcept
{
	return ryml::itoa(buf, (int)v);
}

namespace YAML
{

struct YamlString
{
	std::string yaml;
	YamlString(std::string yamlString);
};

class YamlNodeReader
{
  protected:
	ryml::ConstNodeRef _node;
	bool _invalid;
	std::unordered_map<ryml::csubstr, ryml::id_type>* _index;

	ryml::ConstNodeRef getChildNode(ryml::csubstr key) const;

  public:
	YamlNodeReader();
	YamlNodeReader(const ryml::ConstNodeRef& node);
	YamlNodeReader(const ryml::ConstNodeRef& node, bool useIndex);
	~YamlNodeReader();

	const YamlNodeReader useIndex() const;

	template <typename OutputType>
	OutputType read(ryml::csubstr key) const;
	template <typename OutputType>
	OutputType read(ryml::csubstr key, OutputType defaultValue) const;
	template <typename OutputType>
	OutputType readKey() const;
	template <typename OutputType>
	OutputType readVal() const;
	template <typename OutputType>
	OutputType readVal(OutputType defaultValue) const;
	std::vector<char> readValBase64() const;

	template <typename OutputType>
	bool tryRead(ryml::csubstr key, OutputType& output) const;
	template <typename OutputType>
	bool tryReadKey(OutputType& output) const;
	template <typename OutputType>
	bool tryReadVal(OutputType& output) const;

	size_t childrenCount() const;
	std::vector<YamlNodeReader> children() const;

	bool isValid() const;
	bool isMap() const;
	bool isSeq() const;
	bool hasVal() const;

	const YamlString emit() const;
	void emit(std::string fullFilePath) const;

	const YamlNodeReader operator[](ryml::csubstr key) const;
	const YamlNodeReader operator[](size_t pos) const;
	explicit operator bool() const;
};

class YamlRootNodeReader : public YamlNodeReader
{
private:
	ryml::Tree* _tree;
public:
	YamlRootNodeReader();
	YamlRootNodeReader(std::string fullFilePath, bool onlyInfoHeader);
	YamlRootNodeReader(char* data, size_t size);
	YamlRootNodeReader(const YamlString& yamlString);
	~YamlRootNodeReader();

	/// Returns base class to avoid slicing
	YamlNodeReader sansRoot();
};

class YamlNodeWriter
{
  protected:
	ryml::NodeRef _node;

  public:
	YamlNodeWriter(ryml::NodeRef node);

	/// Conerts writer to a reader
	YamlNodeReader toReader();

	/// Adds an empty child and returns it. Usually a sequence member (no key) that is also a container (no scalar value).
	YamlNodeWriter write();
	/// Adds a child with only a value. Usually a sequence member.
	template <typename InputType>
	YamlNodeWriter write(InputType input);
	/// Adds a child with a key and a value. Usually a mapping member.
	template <typename InputType>
	YamlNodeWriter write(ryml::csubstr key, InputType input);
	/// Adds a child with a key (if the vector is not empty) and a sequence. The callback should specify how to write a vector element to the sequence.
	template <typename InputType, typename Func>
	YamlNodeWriter write(ryml::csubstr key, const std::vector<InputType>& inputVector, Func callback);
	/// Adds a child with only a key. Usually a mapping member that is also a container.
	YamlNodeWriter operator[](ryml::csubstr key);

	YamlNodeWriter writeBase64(ryml::csubstr key, char* data, size_t size);

	/// Marks the current node as a map container
	void setAsMap();
	/// Marks the current node as a sequence container
	void setAsSeq();
	void setFlowStyle();
	void setBlockStyle();

	/// Saves a string into internal buffer. Useful when key strings aren't literals, to extend their lifetime.
	ryml::csubstr saveString(const std::string& string);

	/// Emits a yaml string based on the current node and its subtree
	YamlString emit();
};

class YamlRootNodeWriter : public YamlNodeWriter
{
private:
	ryml::Tree* _tree;
public:
	YamlRootNodeWriter();
	YamlRootNodeWriter(size_t bufferCapacity);
	~YamlRootNodeWriter();

	/// Returns base class to avoid slicing
	YamlNodeWriter sansRoot();
};

/* Template implementations below  */

template <typename OutputType>
OutputType YamlNodeReader::read(ryml::csubstr key) const
{
	OutputType output;
	if (_index)
		_node.tree()->cref(_index->at(key)) >> output;
	else
		_node[key] >> output;
	return output;
}

template <typename OutputType>
OutputType YamlNodeReader::read(ryml::csubstr key, OutputType defaultValue) const
{
	OutputType output;
	if (!tryRead(key, output))
		output = defaultValue;
	return output;
}

template <typename OutputType>
OutputType YamlNodeReader::readKey() const
{
	OutputType output;
	_node >> ryml::key(output);
	return output;
}

template <typename OutputType>
OutputType YamlNodeReader::readVal() const
{
	OutputType output;
	_node >> output;
	return output;
}

template <typename OutputType>
OutputType YamlNodeReader::readVal(OutputType defaultValue) const
{
	OutputType output = defaultValue;
	if (!_invalid)
		_node >> output;
	return output;
}

template <typename OutputType>
bool YamlNodeReader::tryRead(ryml::csubstr key, OutputType& output) const
{
	if (_invalid)
		return false;
	if (!_index)
		return _node.get_if(key, &output);
	if (!_index->count(key))
		return false;
	_node.tree()->cref(_index->at(key)) >> output;
	return true;
}

template <typename OutputType>
bool YamlNodeReader::tryReadKey(OutputType& output) const
{
	if (_invalid)
		return false;
	_node >> ryml::key(output);
	return true;
}

template <typename OutputType>
bool YamlNodeReader::tryReadVal(OutputType& output) const
{
	if (_invalid)
		return false;
	_node >> output;
	return true;
}

template <typename InputType>
inline YamlNodeWriter YamlNodeWriter::write(InputType input)
{
	return YamlNodeWriter(_node.append_child() << input);
}

template <typename InputType>
YamlNodeWriter YamlNodeWriter::write(ryml::csubstr key, InputType input)
{
	return YamlNodeWriter(_node.append_child({ ryml::KEY, key }) << input);
}

template <typename InputType, typename Func>
YamlNodeWriter YamlNodeWriter::write(ryml::csubstr key, const std::vector<InputType>& inputVector, Func callback)
{
	if (inputVector.empty())
		return *this;
	YamlNodeWriter sequenceWriter(_node.append_child({ ryml::KEY, key }));
	sequenceWriter.setAsSeq();
	for (InputType vectorElement : inputVector)
	{
		callback(sequenceWriter, vectorElement);
	}
	return sequenceWriter;
}

}

}
